--[[
   KahLua KonferSK - a suicide kings loot distribution addon.
     WWW: http://kahluamod.com/ksk
     SVN: http://kahluamod.com/svn/konfersk
     IRC: #KahLua on irc.freenode.net
     E-mail: cruciformer@gmail.com
   Please refer to the file LICENSE.txt for the Apache License, Version 2.0.

   Copyright 2008-2010 James Kean Johnston. All rights reserved.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
]]

local K = LibStub:GetLibrary("KKore")
local H = LibStub:GetLibrary("KKoreHash")

if (not K) then
  error ("KahLua KonferSK: could not find KahLua Kore.", 2)
end

local ksk = K:GetAddon ("KKonferSK")
local L = ksk.L
local KUI = ksk.KUI

-- Local aliases for global or Lua library functions
local _G = _G
local tinsert = table.insert
local tremove = table.remove
local setmetatable = setmetatable
local tconcat = table.concat
local tsort = table.sort
local tostring = tostring
local GetTime = GetTime
local min = math.min
local max = math.max
local strfmt = string.format
local strsub = string.sub
local gsub = string.gsub
local strlen = string.len
local strfind = string.find
local strlower = string.lower
local xpcall, pcall = xpcall, pcall
local pairs, next, type = pairs, next, type
local select, assert, loadstring = select, assert, loadstring
local printf = K.printf
local bxor = bit.bxor

local ucolor = K.ucolor
local ecolor = K.ecolor
local icolor = K.icolor
local debug = ksk.debug
local info = ksk.info
local err = ksk.err
local white = ksk.white
local class = ksk.class
local aclass = ksk.aclass

--
-- This file contains general purpose utility functions used throughout KSK.
--

--
-- This function will take a given itemlink and examine its tooltip looking
-- for class restrictions. It will return a class filter mask suitable for
-- use inthe loot system. If no class restriction was found, return the
-- all-inclusive mask.
--
function ksk.GetItemClassFilter (ilink)
  local tnm = GetItemInfo (ilink)
  if (not tnm or tnm == "") then
    return "11111111111", nil
  end

  local tt = ksk.ScanTooltip (ilink)
  local ss = strfmt (ITEM_CLASSES_ALLOWED, "(.-)\n")
  local foo = string.match (tt, ss)
  local boe = nil
  if (string.match (tt, ITEM_BIND_ON_PICKUP)) then
    boe = false
  elseif (string.match (tt, ITEM_BIND_ON_EQUIP)) then
    boe = true
  end

  if (foo) then
    foo = gsub (foo, " ", "")
    local clist = { "0","0","0","0","0","0","0","0","0","1","0" }
    for k,v in pairs ( { string.split (",", foo) } ) do
      local cp = tonumber (K.LClassIndexNSP[v]) or 10
      clist[cp] = "1"
    end
    return tconcat (clist, ""), boe
  else
    return "11111111111", boe
  end
end

--
-- Returns true if the user is thought to be a guild master or officer, false
-- if we cant tell or can tell if they are not.
--
local function try_slow_way (name)
  local found = false
  local n = GetNumGuildMembers ()

  for i = 1, n do
    local nm, _, ri = GetGuildRosterInfo (i)
    if (nm == name) then
      if (ri == 0 or (K.guild.ranks[ri+1].ochatread)) then
        found = true
      end
      break
    end
  end
  return found
end

function ksk:UserIsRanked (name)
  local gn, grn, gri = GetGuildInfo (name)
  if (gn == nil or gri == nil) then
    return try_slow_way (name)
  end
  if (gn ~= K.player.guild) then
    return try_slow_way (name)
  end

  if (gri == 0 or (K.guild.ranks[gri+1].ochatread)) then
    return true
  end
  return false
end

--
-- If we have tethered alts, and teh alt is in the raid, we need to store
-- the UID of the alt's main, else they will not be suicided.
--
function ksk:CreateRaidList (listid)
  local raiders = {}
  for k,v in ipairs (ksk.lists[listid].users) do
    if (ksk:UserIsReserved (v)) then
      tinsert (raiders, v)
    elseif (ksk.raid.users[v]) then
      tinsert (raiders, v)
    elseif (ksk.cfg.tethered and ksk.users[v].alts) then
      for ak,av in pairs (ksk.users[v].alts) do
        if (ksk.raid.users[av]) then
          tinsert (raiders, v)
          break
        end
      end
    end 
  end
  return raiders
end

function ksk:SplitRaidList (raidlist)
  local raiders = {}
  for w in string.gmatch (raidlist, "....") do
    tinsert (raiders, w)
  end
  return raiders
end

--
-- Suicide the user specified by UID on the list LISTID in configuration
-- CFGID. The list of raiders affected by the move is in RLIST. Modifies
-- the list in place. Does not record or transmit events. That is handled
-- elsewhere.
-- Here is how we do suicides. First thing is we find the user being
-- suicided in the raiders list. Any users "above" the user are not
-- subject to moving, so we discount those. Once we have found the user
-- we add their current position to the "movers" list, and add all
-- users below them into the the said movers list. Thus the movers list
-- will contain the full list of people actually affected by the move,
-- and the position they currently occupy in the list. Once we have this
-- it is a simple matter of moving the first user to the bottom of the
-- list and moving everyone else up one slot. The only wrinkle that comes
-- into play is if the user is "frozen". If they are it means that their
-- position in the list needs to remain static, even if they were in the
-- raid. If they were not in the raid they are not affected by moves and
-- will remain static anyway. However, if they are frozen then their
-- position must not change. The easiest way to achieve that is to simply
-- remove any frozen users from the movers list before we suicide the
-- current player, Again, there is a slight exception. If the player being
-- suicided is themselves frozen, we simply pretend that they are not, and
-- move them to the bottom of the list.
--
function ksk:SuicideUserLowLevel (listid, rlist, uid, cfgid, ilink)
  cfgid = cfgid or ksk.currentid
  if (not ksk.configs[cfgid]) then
    return
  end
  if (not ksk.configs[cfgid].lists[listid]) then
    return
  end
  local wl = ksk.configs[cfgid].lists[listid]
  local lu = wl.users
  local found = false
  local movers = {}

  for i = 1, #rlist do
    if (rlist[i] == uid) then
      found = true
      break
    end
  end

  --
  -- This should never happen but hey, whats a simple check.
  --
  if (not found) then
    return
  end

  local foundfirst = false
  for i = 1, #lu do
    found = false
    local j = 1
    while (j <= #rlist) do
      if (lu[i] == rlist[j]) then
        found = true
        break
      end
      j = j + 1
    end

    if (found) then
      if (lu[i] == uid) then
        foundfirst = true
      end
      if (foundfirst) then
        if (lu[i] == uid or not ksk:UserIsFrozen (lu[i], nil, cfgid)) then
          tinsert (movers, i)
        end
      end
    end
  end

  --
  -- This checks to ensure that the first user in the movers list is actually
  -- the userid being suicided. If not, it means that for some reason the user
  -- was not in the list and we therefore have nothing to do. If we only have
  -- one entry in movers, it means the user was not in any actual raid but
  -- was manually suicided by a list admin outside the bounds of a raid. This
  -- is a special case, and in this case the user is moved to the extreme
  -- bottom of the list and everyone except for frozen users is moved up one
  -- slot.
  --
  if (lu[movers[1]] ~= uid) then
    return
  end

  if (#movers == 1) then
    --
    -- Add every user that isn't frozen after the user's current position to
    -- the movers list. This will effectively move the user to the bottom of
    -- the list and move everyone else (that is unfrozen) up one slot. If
    -- there are no users below the user being suicided it means they are
    -- already at the bottom of the list and we have nothing to do.
    --
    local p = movers[1]
    if (p == #lu) then
      return
    end
    for i = p+1, #lu do
      if (not ksk:UserIsFrozen (lu[i], nil, cfgid)) then
        tinsert (movers, i)
      end
    end
  end

  --
  -- Final step. Adjust the actual list users array according to movers.
  -- But first create the undo record.
  --
  local undo = {
    n = #movers, movers = {}, listid = listid, uid = uid, ilink = ilink,
  }
  for i = 1, undo.n do
    undo.movers[i] = lu[movers[i]]
  end
  if (not ksk.csdata[cfgid].undo) then
    ksk.csdata[cfgid].undo = {}
  end
  tinsert (ksk.csdata[cfgid].undo, 1, undo)
  ksk.qf.undobutton:SetEnabled (true)
  local nmove = undo.n - 1
  for i = 1, nmove do
    lu[movers[i]] = lu[movers[i+1]]
  end
  lu[movers[nmove+1]] = uid
  ksk:RefreshLists ()
end

function ksk.AddEvent (cfgid, event, estr, ufn)
  local cfgid = cfgid or ksk.currentid
  debug (9, "adding event %q [%s] for cfgid %s", event, estr, cfgid)

  --
  -- If I am an admin, but not the owner, and I am not syncing yet, do
  -- not even bother adding the event. Its meaningless, as the coadmin
  -- and other users cant sync with us until we have established a sync
  -- relationship.
  --
  local myuid = ksk.csdata[cfgid].myuid
  local ov = ksk.csdata[cfgid].isadmin
  if (ov == 1 and not ksk.configs[cfgid].syncing) then
    return
  end

  local cfg = ksk.configs[cfgid]

  --
  -- This event will change our config checksum. We need to get the CRC for
  -- the event string, and xor it with our current config checksum. We then
  -- need to add it to all co-admins who are currently syncing with us. We
  -- then broadcast the event to the raid or guild. Those syncers that are
  -- on and up to date will perform the action, and remain up to date.
  -- Those that are not online will simply have the event queued for when
  -- they are. Pretty simple really.
  --
  local crc = H:CRC32 (estr)
  local oldsum = cfg.cksum
  local newsum = bxor (oldsum, crc)
  local oldeid = cfg.lastevent
  local eid = ksk:GetEventID (cfgid)
  local scrc = strfmt ("0x%08x", crc)
  cfg.cksum = newsum
  if (ksk.qf.synctopbar) then
    ksk.qf.synctopbar:SetCurrentCRC ()
  end

  if (cfg.syncing) then
    for k,v in pairs (cfg.admins) do
      if (k ~= myuid) then
        if (not v.sync) then
          v.sync = {}
        end
        tinsert (v.sync, strfmt ("%s\8%014.0f\8%s\8%s", event, eid, scrc, estr))
      end
    end
  end

  ksk.CSendAM (cfgid, event, "ALERT", estr, scrc, eid, oldeid, ufn or false)
end

function ksk:RepairDatabases (users, lists)
  if (users == nil) then
    users = true
  end
  if (lists == nil) then
    lists = true
  end

  -- Repair config list count. Also remove sync data added for the admin by
  -- themselves. This means they have co-admins badly configured and they
  -- have an alt on the same account that has written sync data.
  ksk.frdb.nconfigs = 0
  for k,v in pairs (ksk.frdb.configs) do
    ksk.frdb.nconfigs = ksk.frdb.nconfigs + 1
    local to = v.owner
    if (v.admins[to]) then
      if (v.admins[to].sync) then
        v.admins[to].sync = nil
      end
    end
  end

  if (users) then
    for k,v in pairs (ksk.frdb.configs) do
      -- First remove any alts whose main was removed
      for uk, uv in pairs (v.users) do
        if (uv.main and not v.users[uv.main]) then
          ksk:DeleteUser (uk, k, false, true)
        end
      end
      -- Now calculate the correct number of users
      v.nusers = 0
      for uk, uv in pairs (v.users) do
        v.nusers = v.nusers + 1
      end
    end
  end

  if (lists) then
    for k,v in pairs (ksk.frdb.configs) do
      v.nlists = 0
      for lk,lv in pairs (v.lists) do
        v.nlists = v.nlists + 1
        if (not lv.sortorder) then
          lv.sortorder = 1
        end
        if (not lv.def_rank) then
          lv.def_rank = 999
        end
        if (not lv.strictcfilter) then
          lv.strictcfilter = false
        end
        if (not lv.strictrfilter) then
          lv.strictrfilter = false
        end
        if (not lv.extralist) then
          lv.extralist = "0"
        end
        if (not lv.users) then
          lv.users = {}
          lv.nusers = 0
        end
        if (strfind (lv.name, ":")) then
          lv.name = gsub (lv.name, ":", "-")
        end
        local lui = 1
        while (lui <= #lv.users) do
          local uid = lv.users[lui]
          if (not v.users[uid]) then
            tremove (lv.users, lui)
          else
            lui = lui + 1
          end
        end
        lv.nusers = #lv.users
      end
    end
  end
end

function ksk:UpdateDatabaseVersion ()
  local ret = false

  if (not ksk.frdb.dbversion) then
    if (not ksk.frdb.configs) then
      ksk.frdb.dbversion = ksk.dbversion
      return ret
    end

    --
    -- We need to fix a really stupid mistake here. We were supposed to set
    -- the database version if this was a new installation, but we never did.
    -- Now we need to jump through a bunch of hoops to fix that. Eventually
    -- this code will just go away but it will need to stay around for a
    -- good few releases until we can ensure that everyone has updated.
    -- What we have to do is try to determine the actual dbversion that
    -- should have been set, and then let the normal code do its thing.
    -- Note this does not need to be extended, as the bug was fixed at
    -- version 6.
    --
    local _, v = next (ksk.frdb.configs)
    if (v) then
      if (type (v.settings.def_rank) == "string") then
        ksk.frdb.dbversion = 1
      elseif (v.settings.auto_switch ~= nil) then
        ksk.frdb.dbversion = 2
      elseif (v.settings.chat_filter == nil) then
        ksk.frdb.dbversion = 3
      elseif (v.settings.history == nil) then
        ksk.frdb.dbversion = 4
      elseif (v.settings.roll_timeout == nil) then
        ksk.frdb.dbversion = 5
      else
        ksk.frdb.dbversion = 6
      end
    else
      ksk.frdb.dbversion = ksk.dbversion
      ksk.frdb.configs = nil
      return ret
    end
  end

  if (ksk.frdb.dbversion == 1) then
    --
    -- Version 2 corrected the default global guild rank from a string
    -- to a number.
    --
    for k,v in pairs (ksk.frdb.configs) do
      v.settings.def_rank = 999
    end
    ret = true
    ksk.frdb.dbversion = 2
  end

  if (ksk.frdb.dbversion == 2) then
    -- Version 3 removed the auto-switch stuff which was a bad idea.

    for k,v in pairs (ksk.frdb.configs) do
      v.settings.auto_switch = nil
      v.settings.autosw_timeout = nil
      v.settings.auto_extend = nil
      v.settings.final_list = nil
      for kk,vv in pairs (v.lists) do
        vv.nextlist = nil
        vv.bidtimeout = nil
      end
    end
    ret = true
    ksk.frdb.dbversion = 3
  end

  if (ksk.frdb.dbversion == 3) then
    -- Version 4 added chat message filter option

    for k,v in pairs (ksk.frdb.configs) do
      v.settings.chat_filter = true
    end
    ret = true
    ksk.frdb.dbversion = 4
  end

  if (ksk.frdb.dbversion == 4) then
    -- Version 5 added loot history recording, plus remove huser.

    for k,v in pairs (ksk.frdb.configs) do
      v.settings.history = true
      v.settings.huser = nil
      if (not v.history) then
        v.history = {}
      end
    end
    ret = true
    ksk.frdb.dbversion = 5
  end

  if (ksk.frdb.dbversion == 5) then
    -- Version 6 added roll timeouts

    for k,v in pairs (ksk.frdb.configs) do
      v.settings.roll_timeout = 10
      v.settings.roll_extend = 5
    end
    ret = true
    ksk.frdb.dbversion = 6
  end

  if (ksk.frdb.dbversion == 6) then
    -- Version 7 added optional roll before disenchant

    for k,v in pairs (ksk.frdb.configs) do
      v.settings.try_roll = false
    end
    ret = true
    ksk.frdb.dbversion = 7
  end

  if (ksk.frdb.dbversion == 7) then
    --
    -- There was a slight SNAFU if a new co-admin was created with version
    -- 7, where the history table was not being created. Fix that.
    -- There was also a SNAFU where the user count was not being set
    -- correctly when a co-admin created a user.
    --
    for k,v in pairs (ksk.frdb.configs) do
      if (not v.history) then
        v.history = {}
      end
      v.nusers = 0
      for _,_ in pairs (v.users) do
        v.nusers = v.nusers + 1
      end
    end
    ret = true
    ksk.frdb.dbversion = 8
  end

  if (ksk.frdb.dbversion == 8) then
    --
    -- Version 9 added the ability to do fine-grained control over what
    -- gets announced as some whiners said the mod was too spammy.
    --
    for k,v in pairs (ksk.frdb.configs) do
      v.settings.ann_bidchanges = true
      v.settings.ann_winners_raid = true
      v.settings.ann_winners_guild = v.settings.winners
      v.settings.ann_bid_progress = true
      v.settings.ann_bid_usage = true
      v.settings.ann_roll_usage = true
      v.settings.ann_countdown = true
      v.settings.ann_roll_ties = true
      v.settings.ann_cancel = true
      v.settings.ann_no_bids = true
      v.settings.winners = nil
      v.settings.announce = nil
    end
    ksk.frdb.dbversion = 9
    ret = true
  end

  if (ksk.frdb.dbversion <= 9) then
    --
    -- There was a SNAFU with importing users from the original suicide
    -- kings mod, where the class was left blank. We need to fix as much
    -- of that as we can.
    --
    for k,v in pairs (ksk.frdb.configs) do
      local ul = v.users
      for uk,uv in pairs (ul) do
        if (not uv.class or uv.class == "" or not K.IndexClass[uv.class]) then
          local _, pc = UnitClass (uv.name)
          if (pc and pc ~= "" and K.ClassIndex[pc] and K.ClassIndex[pc].u) then
            uv.class = K.ClassIndex[pc]
          else
            err ("forced to remove user %q from configuration %q due to bad Suicide Kings import data.", uv.name, v.name)
            uv.class = K.CLASS_PRIEST
            ksk:DeleteUser (uk, k, true)
          end
        end
      end
    end
  end

  if (ksk.frdb.dbversion == 9) then
    --
    -- Version 10 added the ability to filter offline raiders from the
    -- loot list.
    --
    for k,v in pairs (ksk.frdb.configs) do
      v.settings.hide_absent = false
    end
    ksk.frdb.dbversion = 10
    ret = true
  end

  if (ksk.frdb.dbversion == 10) then
    --
    -- Version 11 added guild rank priorities
    -- 
    for k,v in pairs (ksk.frdb.configs) do
      v.settings.use_ranks = false
      v.settings.rank_prio = {}
    end
    ksk.frdb.dbversion = 11
    ret = true
  end

  if (ksk.frdb.dbversion == 11) then
    --
    -- Version 12 added the ability to controll off-spec rolls and added
    -- suicide rolls, and the ability to set the DE threshold.
    -- 
    for k,v in pairs (ksk.frdb.configs) do
      v.settings.offspec_rolls = true
      v.settings.suicide_rolls = false
      v.settings.bid_threshold = 0
      v.settings.disenchant_below = false
    end
    ksk.frdb.dbversion = 12
    ret = true
  end

  if (ksk.frdb.dbversion == 12) then
    --
    -- Version 13 didnt actual add any new members but it did fix a bunch
    -- of errors in the user removal code. The user may well have a
    -- corrupted user database and we need to fix it here.
    -- This is now done every time the mod initialises.
    --
    ksk:RepairDatabases (true, true)
    ksk.frdb.dbversion = 13
    ret = true
  end

  if (ksk.frdb.dbversion == 13) then
    --
    -- Version 14 added the display missing members option. Also need to
    -- re-repair the user and list databases.
    --
    for k,v in pairs (ksk.frdb.configs) do
      v.settings.ann_missing = true
    end
    ret = true
    ksk:RepairDatabases (true, true)
  end

  --
  -- Fix a potential error with co-admins
  --
  for k,v in pairs (ksk.frdb.configs) do
    for kk,vv in pairs (v.admins) do
      if (vv.active == true) then
        if (vv.lastevent == nil) then
          vv.lastevent = 0
        end
      end
    end
  end

  -- Somehow, tempcfg survives an initial broadcast.
  if (ksk.frdb.defconfig ~= "1") then
    ksk.frdb.tempcfg = nil
  end

  ksk.frdb.dbversion = ksk.dbversion
  return ret
end

--
-- This portion of this file (through to the end) was not written by me.
-- It was a link given to me by Adys on #wowuidev@irc.freenode.net. Many
-- thanks for this code. It has been modified to suit KSK so any bugs are
-- mine.
--

--[[
Copyright (c) Jerome Leclanche. All rights reserved.


Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice, 
       this list of conditions and the following disclaimer.
    
    2. Redistributions in binary form must reproduce the above copyright 
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
--]]

local ksktt = CreateFrame ("GameTooltip", "KSKUTooltip", UIParent,
  "GameTooltipTemplate")
ksktt:SetOwner (UIParent, "ANCHOR_PRESERVE")
ksktt:SetPoint ("CENTER", "UIParent")
ksktt:Hide ()

local function SetTooltipHack (link)
  ksktt:SetOwner (UIParent, "ANCHOR_PRESERVE")
  ksktt:SetHyperlink ("spell:1")
  ksktt:Show ()
  ksktt:SetHyperlink (link)
end

local function UnsetTooltipHack ()
  ksktt:SetOwner (UIParent, "ANCHOR_PRESERVE")
  ksktt:Hide ()
end

function ksk.ScanTooltip (link)
  SetTooltipHack (link)

  local lines = ksktt:NumLines ()
  local tooltiptxt = ""

  for i = 1, lines do
    local left = _G["KSKUTooltipTextLeft"..i]:GetText ()
    local right = _G["KSKUTooltipTextRight"..i]:GetText ()

    if (left) then
      tooltiptxt = tooltiptxt .. left
      if (right) then
        tooltiptxt = tooltiptxt .. "\t" .. right .. "\n"
      else
        tooltiptxt = tooltiptxt .. "\n"
      end
    elseif (right) then
      tooltiptxt = tooltiptxt .. right .. "\n"
    end
  end

  UnsetTooltipHack ()
  return tooltiptxt
end

--[[ NOT CURRENTLY USED
function ksk.GetTooltipLine (link, line, side)
  side = side or "Left"
  SetTooltipHack (link)

  local lines = ksktt:NumLines ()
  if (line > lines) then
    return UnsetTooltipHack()
  end

  local text = _G["KSKUTooltipText"..side..line]:GetText ()
  UnsetTooltipHack ()
  return text
end

function ksk.GetTooltipLines (link, ...)
  local lines = {}
  SetTooltipHack (link)
        
  for k,v in pairs({...}) do
    lines[#lines+1] = _G["KSKUTooltipTextLeft"..v]:GetText()
  end

  UnsetTooltipHack ()
  return unpack (lines)
end
]]

